﻿using Microsoft.DirectX.DirectSound;
using Buffer = Microsoft.DirectX.DirectSound.Buffer;
using System.Threading;
using System.Diagnostics;

namespace TestVoiceNET
{
    abstract class CopyWav
    {
        public CopyWav(System.Windows.Forms.Control aControl)
        {
            m_control = aControl;

            m_waveFormat = new WaveFormat();
            m_waveFormat.FormatTag = WaveFormatTag.Pcm;
            m_waveFormat.SamplesPerSecond = SAMPLE_RATE;
            m_waveFormat.BitsPerSample = SAMPLE_BIT;
            m_waveFormat.Channels = 1;
            m_waveFormat.BlockAlign = (short)(m_waveFormat.Channels * m_waveFormat.BitsPerSample / 8);
            m_waveFormat.AverageBytesPerSecond = m_waveFormat.SamplesPerSecond * m_waveFormat.BlockAlign;

            m_device = new Device();
            m_device.SetCooperativeLevel(m_control, CooperativeLevel.Priority);

            m_secondBuffDesc = new BufferDescription();
            m_secondBuffDesc.Format = m_waveFormat;
            m_secondBuffDesc.BufferBytes = LenWavByte;
            m_secondBuffDesc.LocateInSoftware = true;
            m_secondBuffDesc.ControlPositionNotify = true;
            m_secondBuffDesc.ControlFrequency = true;
            m_secondBuffDesc.ControlPan = true;
            m_secondBuffDesc.ControlVolume = true;
            m_playBuff = new Buffer(m_secondBuffDesc, m_device);

            m_capture = new Capture();
            m_captureBuffDesc = new CaptureBufferDescription();
            m_captureBuffDesc.BufferBytes = LenWavByte;
            m_captureBuffDesc.Format = m_waveFormat;
            m_captureBuffDesc.WaveMapped = true;
            m_captureBuffDesc.ControlEffects = false;
            m_recordBuff = new CaptureBuffer(m_captureBuffDesc, m_capture);

            m_EventPlay = new AutoResetEvent(false);
            m_positionNotifyPlay = new BufferPositionNotify[1];
            m_positionNotifyPlay[0].Offset = LenWavByte - 1;
            m_positionNotifyPlay[0].EventNotifyHandle = m_EventPlay.SafeWaitHandle.DangerousGetHandle();
            m_notifyPlay = new Notify(m_playBuff);
            m_notifyPlay.SetNotificationPositions(m_positionNotifyPlay, 1);

            m_EventRec = new AutoResetEvent(false);
            m_positionNotifyRec = new BufferPositionNotify[1];
            m_positionNotifyRec[0].Offset = LenWavByte - 1;
            m_positionNotifyRec[0].EventNotifyHandle = m_EventRec.SafeWaitHandle.DangerousGetHandle();
            m_notifyRec = new Notify(m_recordBuff);
            m_notifyRec.SetNotificationPositions(m_positionNotifyRec, 1);
        }
        public void Start()
        {
            m_bOpen = true;

            m_threadRec = new Thread(new ThreadStart(Recording));
            m_threadRec.Start();

            m_threadPlay = new Thread(new ThreadStart(Playing));
            m_threadPlay.Start();
        }

        public void Stop()
        {
            m_bOpen = false;

            Break();

            m_recordBuff.Stop();
            m_threadRec.Abort();

            m_threadPlay.Abort();

        }

        public bool IsStart
        {
            get { return m_bOpen; }
        }

        public abstract void Put(WavPack aPack);
	    public abstract void Read(out WavPack aPack);
	    public abstract void Break();

        private void Playing()
        {
            //Debug.WriteLine("Start Playing");

            WavPack Pack;

            while (m_bOpen)
            {
                Read(out Pack);
                if (!m_bOpen) break;

                m_playBuff.Write(0, Pack.Data, LockFlag.None);
                m_playBuff.Play(0, BufferPlayFlags.Default);
            }
        }

        private void Recording()
        {
            //Debug.WriteLine("Start Recording");

            m_recordBuff.Start(true);

            while (true)
            {
                m_EventRec.WaitOne(Timeout.Infinite, true);
                short[] Wav16 = (short[])m_recordBuff.Read(0, typeof(short), LockFlag.None, LenWav);

                Put(new WavPack(Wav16.Length, Wav16));
            }
        }

        public const int SAMPLE_RATE = 8000;
        public const int SAMPLE_BIT = 16;
        public const int LenWavByte = 1920;
        public const int LenWav = LenWavByte/2;

        bool m_bOpen = false;

        private WaveFormat m_waveFormat;

        private Capture m_capture;
        private CaptureBufferDescription m_captureBuffDesc;
        private CaptureBuffer m_recordBuff;
        private AutoResetEvent m_EventRec;
        private BufferPositionNotify[] m_positionNotifyRec;
        private Notify m_notifyRec;
        private Thread m_threadRec;

        private Device m_device;
        private Buffer m_playBuff;
        private BufferDescription m_secondBuffDesc;
        private System.Windows.Forms.Control m_control;
        private Thread m_threadPlay;
        private AutoResetEvent m_EventPlay;
        private BufferPositionNotify[] m_positionNotifyPlay;
        private Notify m_notifyPlay;
    }
}
